/***
 * Checkerboard.java,v 1.4 1997/07/07 20:13:22 dfs Exp
 * Copyright 1997 Original Reusable Objects, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, US
 ***/
package com.oroinc.yaray;

import com.oroinc.math.geometry.*;

/***
 * A checkerboard pigment.  A checkerboard consists of a volume of cubes
 * alternating between two colors.  Each cube touching a face of a
 * cube of the first color is of the second color.
 * <p>
 * Copyright &#169 1996, 1997 Original Reusable Objects, Inc.
 * All rights reserved.
 * <p>
 * @author Daniel F. Savarese
 ***/

public final class Checkerboard implements Pigment {
  /*** The alternating colors of the cube. ***/
  public Color color1, color2;
  /*** The length of the side of a cube. ***/
  public float sideLength;

  /***
   * Creates a checkerboard whose cubes have a sides of length 1.
   * color1 is set to white and color2 is set to black.
   ***/
  public Checkerboard() {
    color1 = new Color(1, 1, 1);
    color2 = new Color(0, 0, 0);
    sideLength = 1;
  }

  /***
   * Creates a checkerboard whose cubes have a sides of length 1.
   * color1 is set to c1 and color2 is set to c2.
   * <p>
   * @param c1  The first color in the checkerboard.
   * @param c2  The second color in the checkerboard.
   ***/
  public Checkerboard(Color c1, Color c2) {
    color1 = new Color(c1);
    color2 = new Color(c2);
    sideLength = 1;
  }


  /***
   * Given the origin of an object, and a point (presumably on the object)
   * return the color (by reference) of that point assuming a Checkerboard
   * pigment.  The checkerboard is a 3 dimensional checkerboard centered
   * on the object whose cubes have side lengths equal to sideLength.
   * <p>
   * @param origin    The origin of the object.
   * @param point     The point on the object.
   * @param colorResult  The color of the gradient at the point.
   ***/
  public void getColor(Point origin, Point point, Color colorResult) {
    GeometricVector v = new GeometricVector(3);
    float x, y, z;
    int i, j, k;

    v = new GeometricVector(Tuple.subtract(point, origin));
    x = v.element(0); y = v.element(1); z = v.element(2);

    // I add ZERO to compensate for block boundaries, where planes and
    // discs will usually be located.  Floating point error will cause
    // intersections to be slightly above and below the integral boundary
    // value, causing speckling on the checker squares.  Adding ZERO
    // pushes the points up a little to eliminate speckling.  This works
    // because we usually don't define planes and discs and such on
    // coordinates with several decimal places.  Likewise for checkerboard
    // side lengths.

    i = (int)Math.floor(x/sideLength + Util.ZERO);
    j = (int)Math.floor(y/sideLength + Util.ZERO);
    k = (int)Math.floor(z/sideLength + Util.ZERO);

    if(Util.isEven(i + j + k))
      colorResult.assign(color1);
    else
      colorResult.assign(color2);
  }
    
}
