/***
 * Disc.java,v 1.4 1997/07/07 20:13:22 dfs Exp
 * Copyright 1997 Original Reusable Objects, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, US
 ***/
package com.oroinc.yaray;

import com.oroinc.math.geometry.*;

/***
 * A Disc WorldObject.
 * <p>
 * Copyright &#169 1996, 1997 Original Reusable Objects, Inc.
 * All rights reserved.
 * <p>
 * @author Daniel F. Savarese
 ***/

public class Disc extends WorldObject {
  /*** The radius of the disc. ***/
  private float __radius;
  /*** The center of the disc. ***/
  private Point __center;
  /*** The normal of the disc surface. ***/
  private GeometricVector __normal;

  /***
   * Creates a Disc object with the given characteristics.
   * <p>
   * @param normal  The normal of the disc surface.
   * @param center  The center of the disc.
   * @param radius  The radius of the disc.
   * @param pigment The pigment of the disc.
   * @param finish  The finish of the disc.
   ***/
  public Disc(GeometricVector normal, Point center, float radius,
	      Pigment pigment, Finish finish) {
    super(pigment, finish);
    __radius = radius;
    __center = new Point(center);
    __normal = new GeometricVector(normal);
  }

  /***
   * Determines if ray intersects disc.  If it does, checks that
   * point of intersection is closer to origin of ray than any previous
   * point of intersection.  If it is closer, returns true, if not,
   * returns false.
   * <p>
   * @param ray  Ray to be tested for intersection.
   * @return true If ray intersects disc and is closer to origin than any
   *              previous point of intersection, false otherwise.
   ***/
  public boolean intersect(IntersectionRay ray) {
    float t;
    GeometricVector v;

    if((t = GeometricVector.dot(__normal, ray.direction)) == 0)
      return false;

    t = -GeometricVector.dot(__normal, Tuple.subtract(ray.origin, __center))/t;

    if(t < Util.ZERO)
      return false;

    v = new GeometricVector(Tuple.add(ray.origin, ray.direction.multiply(t)));
    v.subtractFrom(__center);

    if(v.magnitude() > __radius)
      return false;
    
    if(t < ray.extent) {
      ray.extent = t;
      ray.intersection = this;
      return true;
    }
		
    return false;
  }

  /***
   * Given a world description and the ray that intersects the disc, returns
   * the color of the disc at the point of intersection.
   * <p>
   * @param world  The world containing the disc.
   * @param intersection  The current ray being traced.
   * @param colorResult  The color at the point of intersection if any.
   ***/


  public void getColor(World world, Ray intersection, Color colorResult) {
    int numLights;
    Point point;
    Color newColor;

    point = intersection.endpoint();
    numLights = world.lights.length;
    _pigment_.getColor(__center, point, colorResult);

    newColor = _finish_.ambientColor(colorResult);
    newColor.addTo(_finish_.reflectionColor(world, __normal, intersection));
    newColor.addTo(_finish_.refractionColor(world, __normal, intersection));

    while(numLights-- > 0)
      newColor.addTo(
      		     _finish_.lightColor(world, numLights,
					 __normal, intersection, colorResult)
		     );

    colorResult.assign(newColor);
  }
}


