/***
 * Finish.java,v 1.4 1997/07/07 20:13:22 dfs Exp
 * Copyright 1997 Original Reusable Objects, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, US
 ***/
package com.oroinc.yaray;

import com.oroinc.math.geometry.*;

/***
 * The finish of a surface.
 * <p>
 * Copyright &#169 1996, 1997 Original Reusable Objects, Inc.
 * All rights reserved.
 * <p>
 * @author Daniel F. Savarese
 ***/

public class Finish {
  public float kAmbient, kDiffuse, kSpecular, kReflectivity;
  public float kTransmitted, nRefraction;
  public float brilliance, roughness;

  public Finish() {
    kAmbient = kDiffuse = kSpecular = kReflectivity =
      kTransmitted = nRefraction = brilliance = roughness = 0;
  }

  public Color ambientColor(Color color) {
    Color result = new Color(color);
    result.multiplyBy(kAmbient);
    return result;
  }


  /***
   * Returns the specular and diffuse reflection components of the finish
   * color.
   * <p>
   * @param world A pointer to the world description.
   * @param lightIndex  The index into the world.lights[] array
   *       indicating the light for which the color is being computed.
   * @param normal The surface normal at the point where the
   *               color is being computed.
   * @param intersection The ray that intersects the object defining where
   *                     the color is being computed.
   * @param color The base color of the object at the intersection point.
   ***/

   public Color lightColor(World world, int lightIndex, GeometricVector normal,
			   Ray intersection, Color color) {
     float number;
     Light light;
     Color lightColor = new Color();
     Point point;
     GeometricVector lightVector,
                     halfway     = new GeometricVector(3),
                     viewer;
     Color diffuse = new Color(), specular = new Color();
     IntersectionRay ray = new IntersectionRay();

     point = intersection.endpoint();
     light = world.lights[lightIndex];
     lightVector = new GeometricVector(Tuple.subtract(light.location, point));
     halfway.assign(lightVector);  // use halfway as temporary
     lightVector.normalize();

     if((number = GeometricVector.dot(normal, lightVector)) > 0){
       // The angle of incidence is such that there is a reflection
       ray.extent = halfway.magnitude();
       ray.origin.assign(point);
       ray.direction.assign(lightVector);

       // Check for intersections
       world.raytrace(ray, lightColor);

       // Yes an ugly goto.  If the ray to the light source intersects
       // some object before getting to the light source, the specular
       // and diffuse components are zero.
       if(ray.intersection == null)
	 lightColor.assign(light.color);
       else if(ray.intersection._finish_.kTransmitted == 0) {
       // The angle of incidence is such that there is no reflection
	 //diffuse.zero();
	 //specular.zero();
	 return new Color(0, 0, 0);
       }

    // Compute diffuse component
       diffuse = color.multiply(lightColor);

       diffuse.multiplyBy((float)(kDiffuse * Math.pow(number, brilliance)));

       viewer = new GeometricVector(Tuple.subtract(world.eye, point));
       //    halfway = lightVector + viewer.Normalized();
       viewer.normalize();
       halfway = new GeometricVector(Tuple.add(lightVector, viewer));

    // Compute specular component if the magnitude of the halfway is
    // greater than some very small number close to zero.
       if(halfway.magnitude() < World.EPSILON)
	 specular.zero();
       else {
	 halfway.normalize();
	 if(roughness != 0 &&
	    (number = GeometricVector.dot(normal, halfway)) > 0){
	   specular.assign(lightColor);
	   specular.multiplyBy((float)
			       (kSpecular * Math.pow(number, 1/roughness)));
	 }
	 else
	   specular.zero();
       }
     } else {
       // The angle of incidence is such that there is no reflection
       //diffuse.zero();
       //specular.zero();
       return new Color(0, 0, 0);
     }

     return new Color(Tuple.add(diffuse, specular));
   }


  /***
   * Returns the pure reflection component of the finish color.
   * <p>
   * @param world  A pointer to the world description.
   * @param normal The surface normal at the point where the
   *               color is being computed.
   * @param intersection  The ray that intersects the object defining where
   *                      the color is being computed.
   ***/

  public Color reflectionColor(World world, GeometricVector normal,
			       Ray intersection) {
    float number;
    Color reflective = new Color();
    IntersectionRay ray;

    if(kReflectivity > 0 &&
       (number = GeometricVector.dot(normal, intersection.direction)) != 0) {
      ray = new IntersectionRay();
      // The object is reflective and the angle of incidence of the ray
      // is such that there is a reflection.
      ray.extent = Util.INFINITY;
      ray.origin = intersection.endpoint();

      if(number < 0) {
	ray.direction = new GeometricVector(normal.multiply(-2*number));
	ray.direction.addTo(intersection.direction);
      } else {
	ray.direction = new GeometricVector(normal.multiply(2*number));
	ray.direction.subtractFrom(intersection.direction);
      }

      world.raytrace(ray, reflective);

      reflective.multiplyBy(kReflectivity);
    } else
      reflective.zero();    // No reflection

    return reflective;
  }


  /***
   * Returns the pure refraction component of the finish color.
   * <p>
   * @param world   A pointer to the world description.
   * @param normal  The surface normal at the point where the
   *                color is being computed.
   * @param intersection  The ray that intersects the object defining where
   *                      the color is being computed.
   ***/

  public Color refractionColor(World world, GeometricVector normal,
			       Ray intersection) {
    IntersectionRay ray;
    Color color = new Color();
    float determinant, eta, ni;
    GeometricVector newNormal;

    if(kTransmitted > 0){  // Is the object refractive?
      ray = new IntersectionRay();
      ray.extent = Util.INFINITY;
      ray.origin = intersection.endpoint();
      ray.origin.addTo(intersection.direction.multiply(Util.ZERO));

      // We assume that the index of refraction of the intervening
      // medium is 1.0.  Transition is always considered from air into
      // object since we are assuming hollow objects with infinitely
      // thin surfaces.
      eta = 1/nRefraction;

      ni = GeometricVector.dot(normal, intersection.direction);

      if(ni <= 0){
	ni = -ni;
	newNormal = new GeometricVector(normal);
      }
      else
	newNormal = new GeometricVector(normal.inverse());

      if((determinant = 1 + eta*eta*(ni*ni - 1)) >= 0){
	determinant = (float)(eta*ni - Math.sqrt(determinant));

	ray.direction.assign(newNormal);
	ray.direction.multiplyBy(determinant);
	ray.direction.addTo(intersection.direction.multiply(eta));
	ray.direction.normalize();

	// Get color of refracted light
	world.raytrace(ray, color);

	color.multiplyBy(kTransmitted);

	return color;
      }

      // Internal Reflection occurs
      if(ni < 0) {
	ray.direction.assign(normal);
	ray.direction.multiplyBy(-2*ni);
	ray.direction.addTo(intersection.direction);
      } else {
	ray.direction.assign(normal);
	ray.direction.multiplyBy(2*ni);
	ray.direction.subtractFrom(intersection.direction);
      }

      // Get color of reflected light
      world.raytrace(ray, color);
    
      color.multiplyBy(1 - kReflectivity);
      return color;
    }

    color.zero();

    // Object is non-refractive therefore refractive component is black.
    return color;
  }
}
