/***
 * Gradient.java,v 1.4 1997/07/07 20:13:22 dfs Exp
 * Copyright 1997 Original Reusable Objects, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, US
 ***/
package com.oroinc.yaray;

import com.oroinc.math.geometry.*;

/***
 * A gradient pigment.
 * <p>
 * Copyright &#169 1996, 1997 Original Reusable Objects, Inc.
 * All rights reserved.
 * <p>
 * @author Daniel F. Savarese
 ***/

public final class Gradient implements Pigment {
  public Color color1, color2;
  public GeometricVector gradient;

  /*** Creates a default gradient from black to white. ***/
  public Gradient() {
    color1 = new Color(0, 0, 0);
    color2 = new Color(1, 1, 1);
    gradient = new GeometricVector(0, 0, 1);
  }


  public Gradient(Color c1, Color c2, GeometricVector g) {
    color1 = new Color(c1);
    color2 = new Color(c2);
    gradient = new GeometricVector(g);
  }

  /***
   * Given the origin of an object, and a point (presumably on the object)
   * return the color (by reference) of that point assuming a gradient
   * pigment.  The gradient is defined as follows.  Project the
   * point (i.e. the vector point - origin) onto the gradient vector.  At the
   * origin of the gradient vector the color is color1, and at the end of
   * the gradient vector, the color is color2.  The colors change
   * from color1 to color2 and then again from color2 to color1, etc.
   * as you travel along the gradient vector and beyond it's two extremes.
   * <p>
   * After projecting the point onto the gradient, the returned color
   * is the color on the gradient at the projected point.  You can
   * do the color interpolation in various ways.  This particular
   * implementation uses a linear interpolation.  I've commented
   * out an alternate method that uses a cosine function providing
   * wider bands near the endpoints and sharper gradation near the
   * center.
   * <p>
   * @param origin    The origin of the object.
   * @param point     The point on the object.
   * @param colorResult  The color of the gradient at the point.
   ***/

  public void getColor(Point origin, Point point, Color colorResult) {
    int floor;
    float proportion;

    // Project point onto line from origin in direction of gradient
    // proportion = (Dot(point - origin, gradient)/Dot(gradient, gradient));
    proportion = GeometricVector.dot(Tuple.subtract(point, origin), gradient);
    proportion/=(GeometricVector.dot(gradient, gradient));

    floor = (int)Math.floor(proportion);
    proportion-=floor;

    // Account for the gradient from color2 back to color1
    if(Util.isOdd(floor))
      proportion = 1 - proportion;

    // Alternate interpolation method.  Just uncomment to implement.
    proportion =
      (float)(0.5*(Math.cos((1 + proportion)*Math.PI) + 1));

    // colorResult = proportion*color2 + (1 - proportion)*color1;
    colorResult.assign(color2);
    colorResult.multiplyBy(proportion);
    colorResult.addTo(color1.multiply(1 - proportion));
  }

}




