/***
 * Sphere.java,v 1.4 1997/07/07 20:13:24 dfs Exp
 * Copyright 1997 Original Reusable Objects, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. 
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, US
 ***/
package com.oroinc.yaray;

import com.oroinc.math.geometry.*;

/***
 * A sphere.
 * <p>
 * Copyright &#169 1996, 1997 Original Reusable Objects, Inc.
 * All rights reserved.
 * <p>
 * @author Daniel F. Savarese
 ***/

public class Sphere extends WorldObject {
  private float __radius;
  private Point __center;

  public Sphere(Point center, float radius, Pigment pigment, Finish finish) {
    super(pigment, finish);
    __radius = radius;
    __center = new Point(center);
  }

  /***
   * Determines if ray intersects sphere.  If it does, checks that
   * point of intersection is closer to origin of ray than any previous
   * point of intersection.  If it is closer, returns true, if not,
   * returns false.
   * <p>
   * @param ray  Ray to be tested for intersection.
   * @return true If ray intersects disc and is closer to origin than any
   *              previous point of intersection, false otherwise.
   ***/
  public boolean intersect(IntersectionRay ray) {
    GeometricVector u;
    float a, b, c, d, t, t1, t2;

    u = new GeometricVector(Tuple.subtract(ray.origin, __center));
    a = GeometricVector.dot(ray.direction, ray.direction);
    b = 2*GeometricVector.dot(u, ray.direction);
    c = GeometricVector.dot(u, u) - __radius*__radius;
    d = b*b - 4*a*c;

    if(d >= 0){
      t1 = 0.5f*(-b + (float)Math.sqrt(d))/a;
      t2 = 0.5f*(-b - (float)Math.sqrt(d))/a;
      t = Math.min(t1, t2);
      if(t < 0)
	t = Math.max(t1, t2);

      if(t > Util.ZERO && t < ray.extent) {
	ray.extent = t;
	ray.intersection = this;
	return true;
      }
    }

    return false;
  }

  /***
   * Given a world description and the ray that intersects the sphere, returns
   * the color of the sphere at the point of intersection.
   * <p>
   * @param world         The world containing the sphere.
   * @param intersection  The current ray being traced.
   * @param colorResult   The color at the point of intersection if any.
   ***/

  public void getColor(World world, Ray intersection, Color colorResult) {
    int numLights;
    Point point;
    GeometricVector normal = new GeometricVector(3);
    Color newColor;

    point = intersection.endpoint();
    normal = new GeometricVector(Tuple.subtract(point, __center));
    normal.normalize();
    numLights = world.lights.length;
    _pigment_.getColor(__center, point, colorResult);

    newColor = _finish_.ambientColor(colorResult);
    newColor.addTo(_finish_.reflectionColor(world, normal, intersection));
    newColor.addTo(_finish_.refractionColor(world, normal, intersection));

    while(numLights-- > 0)
      newColor.addTo(
		     _finish_.lightColor(world, numLights, normal,
					 intersection, colorResult)
		     );

    colorResult.assign(newColor);
  }
}

